#!/usr/bin/env python

"""
interfaces.py - user interface components for PEEQ

Copyright (C) 2004-2005, David Boddie

This file is part of PEEQ.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""

from qt import QBrush, QEvent, QImage, QPainter, QPixmap, QScrollView, QSize, \
               QSizePolicy, QTimer, Qt, \
               QVBox, QWidget, PYSIGNAL, SIGNAL


thumbnail_width = 50
thumbnail_height = 50

class WidgetBox(QScrollView):

    # Redefine the tr() function for this class.
    def tr(self, text):
    
        return qApp.translate("CapturedWidgets", text)
    
    def __init__(self, parent = None):
    
        self.widgets = {}
        
        QScrollView.__init__(self, parent)
        self.layout_manager = QVBox(self.viewport())
        self.layout_manager.setSizePolicy(
            QSizePolicy.Fixed, QSizePolicy.Expanding
            )
        self.addChild(self.layout_manager)
    
    def eventFilter(self, obj, event):
    
        if obj in self.widgets.keys():
        
            if event.type() == QEvent.Close and \
                obj.getWFlags() & Qt.WDestructiveClose:
            
                self.remove_widget(obj)
                self.update_area()
            
            elif event.type() == QEvent.ChildInserted or \
                event.type() == QEvent.ChildRemoved or \
                event.type() == QEvent.Resize or \
                event.type() == QEvent.Show:
            
                self.widgets[obj].update_picture()
        
        return 0
    
    def add_widgets(self, widgets):
    
        for widget in widgets:
        
            widget.installEventFilter(self)
            self.connect(widget, SIGNAL("destroyed()"), self.remove_widget)
            
            thumbnail = CapturedWidget(self.layout_manager, widget)
            self.connect(thumbnail, PYSIGNAL("deleted"), self.remove_thumbnail)
            thumbnail.show()
            self.widgets[widget] = thumbnail
        
        self.update_area()
    
    def remove_widget(self, widget = None):
    
        if widget is None:
        
            widget = self.sender()
        
        if self.widgets.has_key(widget):
        
            self.widgets[widget].close()
            del self.widgets[widget]
            self.update_area()
    
    def remove_thumbnail(self, widget):
    
        del self.widgets[widget]
        self.update_area()
    
    def update_area(self):
    
        self.layout_manager.setMinimumSize(
            thumbnail_width, len(self.widgets.keys())*thumbnail_height
            )
        self.layout_manager.setMaximumSize(
            thumbnail_width, len(self.widgets.keys())*thumbnail_height
            )
    
    def sizeHint(self):
    
        return QSize(thumbnail_width, self.height())


class CapturedWidget(QWidget):

    def __init__(self, parent, widget):
    
        QWidget.__init__(self, parent)
        
        self.brush = QBrush(Qt.black)
        self.widget = widget
        
        self.update_picture()
        QTimer.singleShot(5000, self.update_picture)
    
    def update_picture(self):
    
        try:
        
            # Create a picture of the widget's contents.
            image = QPixmap.grabWidget(self.widget).convertToImage()
        
        except RuntimeError:
        
            self.emit(PYSIGNAL("deleted"), (self.widget,))
            self.close(1)
            return
        
        # Scale the pixmap to a default size.
        image = image.smoothScale(
            thumbnail_width, thumbnail_height, QImage.ScaleMin
            )
        
        self.pixmap = QPixmap(image)
        self.update()
    
    def sizeHint(self):
    
        return QSize(thumbnail_width, thumbnail_height)
    
    def minimumSizeHint(self):
    
        return QSize(thumbnail_width, thumbnail_height)
    
    def paintEvent(self, event):
    
        painter = QPainter()
        painter.begin(self)
        painter.fillRect(0, 0, thumbnail_width, thumbnail_height, self.brush)
        painter.drawPixmap(
            int(thumbnail_width/2 - self.pixmap.width()/2),
            int(thumbnail_height/2 - self.pixmap.height()/2), self.pixmap
            )
        painter.end()
    
    def mousePressEvent(self, event):
    
        try:
        
            self.widget.show()
            self.widget.raiseW()
        
        except RuntimeError:
        
            self.emit(PYSIGNAL("deleted"), (self.widget,))
            self.close(1)
