#!/usr/bin/env python

"""
peeq.py - a Python editing environment for Qt

Copyright (C) 2004-2005, David Boddie

This file is part of PEEQ.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""

import imp, os, string, sys, traceback, types

from PEEQ.formatters import encode
from PEEQ.interfaces import WidgetBox
from PEEQ.iobuffer import Buffer, History

from qt import QAction, QApplication, QComboBox, QDialog, QEvent, \
               QFileDialog, QFont, \
               QHBoxLayout, \
               QInputDialog, QKeySequence, QLineEdit, QListBox, QMainWindow, \
               QMessageBox, QMouseEvent, \
               QObject, QPoint, \
               QPopupMenu, QPushButton, QStatusBar, QTextBrowser, QTextEdit, \
               QThread, Qt, \
               QVBox, QVBoxLayout, QWidget, PYSIGNAL, SIGNAL, SLOT, qApp

try:

    import qtext
    use_qtext = 1

except ImportError:

    use_qtext = 0


# Import the __future__ module to enable code using future language changes
# to be compiled interactively.

import __future__


# Define the application details.

appname = "peeq"
__version__ = "0.11"


import __builtin__, keyword


class MatchError(Exception):

    pass

class Completer:

    extra_attributes = ("__class__", "__base__", "__module__", "__bases__")
    
    def __init__(self, module):
    
        self.module = module
        self.match_try = 0
        self.maximum_tries = 2
    
    def find_object(self, text, delimiters = "(, "):
    
        # Find the maximum index to start at. If no starting indices are found
        # then -1 is used. This is fine because we actually want to start
        # scanning the text at the beginning of the string in that situation.
        indices = map(
            lambda delimiter: text.rfind(delimiter), delimiters
            )
        
        start_at = max(indices) + 1
        return start_at, text[start_at:]
    
    def descend_name(self, pieces, namespaces):
    
        if pieces == []:
        
            return None
        
        piece = pieces.pop(0).strip()
        
        if pieces == []:
        
            matches = []
            
            for namespace in namespaces:
            
                matches = matches + filter(
                    lambda name: name.find(piece) == 0, dir(namespace)
                    )
            
            if matches == []:
            
                matches = matches + filter(
                    lambda name: name.find(piece) == 0, keyword.kwlist
                    )
            
            return matches
        
        else:
        
            matches = []
            
            for namespace in namespaces:
            
                if piece in dir(namespace):
                
                    obj = getattr(namespace, piece)
                    
                    matches = matches + map(
                        lambda name: piece+"."+name,
                        self.descend_name(pieces, (obj,))
                        )
            
            return matches
    
    def extend_attributes(self, obj):
    
        dict = {}
        for attribute in self.extra_attributes:
        
            if hasattr(obj, attribute):
                dict.update({attribute: getattr(obj, attribute)})
        
        return dict
    
    def match(self, text, cursor):
    
        # Work backwards from the cursor looking for the start of object names.
        start_at, full_name = self.find_object(text[:cursor])
        
        namespaces = (self.module, __builtin__)
        
        try:
        
            return start_at, self.descend_name(full_name.split("."), namespaces)
        
        except MatchError:
        
            return len(text), []


# Define an object to contain the license for this application.
class License:

    def __init__(self, text):
    
        self.name = text[1:].split(" ")[0]
        lines = text.split("\n")
        lines[1] = "<h2>%s</h2>" % lines[1]
        lines[3] = "<h3>%s</h3>" % lines[3]
        lines.insert(5, "<pre>")
        lines.append("</pre>")
        self.text = "\n".join(lines)
    
    def __repr__(self):
    
        self.browser = QTextBrowser()
        self.browser.setTextFormat(QTextBrowser.RichText)
        self.browser.setText(self.text)
        self.browser.show()
        return ""

__license__ = License(__doc__)

# Define a start up message for the interpreter.

start_up = \
"""
<h1 class = "Title">%s %s</h1>

<p>
  <em>python:</em>
  Type "help", "copyright", "credits" or "license" for more information.
</p>
<p>
  <em>%s:</em>
  Type "%s_license" for license information specific to this environment.
</p>
""" % (appname, __version__, appname, __license__.name)

class NoValueClass:

    pass

NoValue = NoValueClass()


class Traceback:

    def __init__(self, text):
    
        self.text = text


# Define a docstring for the user module.
user_docstring = __doc__[1:string.find(__doc__, "\n", 1)]


class Environment(QObject):

# Define a function to obtain the class attributes for an instance.

    adir_function = (
        'def adir(obj):\n'
        '\n'
        '    if hasattr(obj, "__class__") and not hasattr(obj, "__bases__"):\n'
        '    \n'
        '        return obj.__class__.__dict__.keys()\n'
        '    \n'
        '    else:\n'
        '    \n'
        '        return []\n'
        )

    def __init__(self):
    
        QObject.__init__(self)
        
        # Import some modules to customise the environment.
        self.setup_environment()
        
        # Delayed code for execution.
        self.delayed = []
    
    def setup_environment(self):
    
        # Construct a module to use to contain Python executed by the
        # user.
        self.module = imp.new_module("__user__")
        
        # Add a docstring for this module.
        self.module.__doc__ = user_docstring
        
        # Copy the sys module into the user's environment.
        self.module.sys = sys
        
        # Copy the qt module into the user's environment.
        import qt
        self.module.qt = qt
        
        # Define some functions for the user.
        
        # Introduce a function to return the class attributes of instance
        # objects.
        
        adir = compile(self.adir_function, "<setup_functions>", "exec")
        exec adir in self.module.__dict__
                
        # Set up an object containing the license information in the
        # user module.
        self.module.peeq_license = __license__
        #self.module.LICENSE = __doc__[1:]
    
    def execute(self, command_line, use_delayed = 1):
    
        if len(command_line) == 0:
        
            return NoValue
        
        if use_delayed:
        
            if command_line[0] in " \t":
            
                # Add the line to the list of delayed command lines.
                self.delayed.append(command_line)
                return NoValue
            
            if self.delayed != []:
            
                command_lines = "".join(self.delayed) + command_line
            
            else:
            
                command_lines = command_line
        
        try:
        
            value = eval(command_lines, self.module.__dict__)
            
            # Set the _ variable to the value returned.
            self.module._ = value
            
            if use_delayed:
            
                # The pending input was executed successfully so clear
                # the delayed command lines list.
                self.delayed = []
            
            return value
        
        except SyntaxError:
        
            # Catch Syntax Errors and pass them on for another attempt
            # to compile and execute them.
            pass
        
        except:
        
            # Catch all other errors and report them.
            
            if use_delayed:
            
                # Clear the delayed command line list.
                self.delayed = []
            
            return self.format_exc()
        
        
        # Try again with the exec command.
        
        # Construct a flags word for any features imported from the
        # __future__ module. Peculiarly, this has to be done because
        # the compile function appears not to be aware of any __future__
        # imports done from the command line.
        
        flags = 0
        user_values = filter(
            lambda x: isinstance(x, __future__._Feature),
            self.module.__dict__.values()
            )
        
        for value in __future__.__dict__.values():
        
            if isinstance(value, __future__._Feature):
            
                if value in user_values:
                
                    flags = flags | value.compiler_flag
        
        try:
        
            # Compile the code using the flags determined from the imported
            # features. The last parameter indicates that these flags should
            # be taken into account when compiling.
            
            code = compile(command_lines, "__user__", "exec", flags, 0)
            exec code in self.module.__dict__
            
            if use_delayed:
            
                # The pending input was executed successfully so clear
                # the delayed command lines list.
                self.delayed = []
            
            # No values are returned from the execution of code in this
            # manner unless we are using existing lines from the delayed
            # lines list.
            return NoValue
        
        except SyntaxError:
        
            msg = sys.exc_info()[1].msg
            #print encode(msg)
            
            if string.find( msg, "unexpected EOF while parsing" ) != -1:
            
                if use_delayed:
                
                    # Delay code for later execution.
                    self.delayed.append(command_line)
                
                return NoValue
            
            elif string.find(msg, "expected an indented block") != -1:

                # Only take this seriously if there is no preceding text
                # or if these last two lines have the same indentation.
                
                if use_delayed and self.delayed != [] and \
                    ( self.indentation(self.delayed[-1]) != \
                      self.indentation(command_line)          ):
                
                    # Keep the line for later if there is preceding text and
                    # the indentation has changed.
                    
                    self.delayed.append(command_line)
                    return NoValue
            
            # An actual Syntax Error.
            
            if use_delayed:
            
                # Clear the delayed command line list.
                self.delayed = []
            
            return self.format_exc()
        
        except:
        
            # Other errors are caught here.
            
            if use_delayed:
            
                # Clear the delayed command line list.
                self.delayed = []
            
            return self.format_exc()
    
    def format_exc(self):
    
        lines = map(
            lambda line: line,
            #traceback.format_exception_only(*(sys.exc_info()[:2]))
            traceback.format_exception(*(sys.exc_info()))
            )
        
        return Traceback("\n".join(lines))
    
    def indentation(self, s, tabsize = 4):
    
        l = 0
        
        for i in s:
        
            if i == " ":
                l = l + 1
            elif i == "\t":
                l = l + tabsize
            else:
                break
        
        return l


class Editor(QWidget):

    # Redefine the tr() function for this class.
    def tr(self, text):
    
        return qApp.translate("Editor", text)
    
    def __init__(self, environment):
    
        QWidget.__init__(self)
        
        self.environment = environment
        
        if use_qtext:
            self.text_edit = qtext.QextScintilla(self)
            self.text_edit.setLexer(qtext.QextScintillaLexerPython(self))
        else:
            self.text_edit = QTextEdit(self)
        
        font = QFont()
        font.setFamily("Courier")
        self.text_edit.setFont(font)
        
        self.run_button = QPushButton(self.tr("Run"), self)
        self.close_button = QPushButton(self.tr("Close"), self)
        
        self.status_bar = QStatusBar(self)
        
        self.connect(self.run_button, SIGNAL("clicked()"), self.run_script)
        self.connect(
            self.close_button, SIGNAL("clicked()"), self, SLOT("close()")
            )
        
        panel_layout = QHBoxLayout()
        panel_layout.addStretch(1)
        panel_layout.addWidget(self.run_button)
        panel_layout.addSpacing(4)
        panel_layout.addWidget(self.close_button)
        
        layout = QVBoxLayout(self)
        layout.addSpacing(4)
        layout.addWidget(self.text_edit, 1)
        layout.addSpacing(4)
        layout.addLayout(panel_layout)
        layout.addWidget(self.status_bar)
    
    def run_script(self):
    
        # Examine the text submitted.

        # Submit it to the user's running environment.
        try:

            # Retrieve a string value from the environment's execute method.
            text = str(self.text_edit.text())

            if text[-1:] != ["\n"]:

                text = text + "\n"

            value = self.environment.execute(text)

            if isinstance(value, Traceback):

                self.emit(PYSIGNAL("writeStdOut"), (value.text + "\n",))
                last_line = filter(
                    lambda line: line != "", value.text.split("\n")
                    )[-1]

            elif value is not NoValue:# and value is not None:

                self.emit(PYSIGNAL("writeStdOut"), (repr(value) + "\n",))
                last_line = repr(value).split("\n")[-1]
            
            else:
            
                last_line = "OK"
            
            self.status_bar.message(last_line, 2000)
        
        except SyntaxError:
        
            pass


class LineEdit(QLineEdit):

    def __init__(self, parent):
    
        QLineEdit.__init__(self, parent)


class GUIInterface(QMainWindow):

    """GUIInterface(QMainWindow)
    
    This class handles the interactions between a CommandLine object and an
    Environment object.
    
    gui = GUIInterface(application, environment)
    
    The GUIInterface object is a subclass of QObject in order to allow
    executed Python code to access Qt classes.
    """
    
    display_width = 80
    editor_keys = (Qt.Key_Left, Qt.Key_Right, Qt.Key_Up, Qt.Key_Down,
                   Qt.Key_Backspace, Qt.Key_Delete, Qt.Key_Home, Qt.Key_End)
    documents_path = os.path.join(os.getenv("QTDIR"), "doc", "html")
    
    # Redefine the tr() function for this class.
    def tr(self, text):
    
        return qApp.translate("GUIInterface", text)
    
    def __init__(self, app):
    
        QMainWindow.__init__(self)
        
        self.app = app
        
        central_widget = QWidget(self)
        
        left_frame = QWidget(central_widget)
        
        self.text_browser = QTextBrowser(left_frame)
        self.text_browser.setFocusPolicy(QMainWindow.NoFocus)
        self.text_browser.setWordWrap(QTextEdit.FixedColumnWidth)
        self.text_browser.setTextFormat(QTextEdit.RichText)
        
        self.combo_edit = QComboBox(left_frame)
        self.combo_edit.setAutoCompletion(1)
        self.combo_edit.setEditable(1)
        self.combo_edit.installEventFilter(self)
        self.combo_edit.lineEdit().installEventFilter(self)
        self.combo_edit.setDuplicatesEnabled(0)
        
        self.captured_widgets = WidgetBox(central_widget)
        self.captured_widgets.hide()
        
        # Setup the environment.
        self.setup_environment()
        
        self.setup_menus()
        
        self.connect(
            self.combo_edit, SIGNAL("activated(const QString &)"),
            self.combo_edit, SLOT("setEditText(const QString &)")
            )
        self.connect(
            self.environment, PYSIGNAL("writeStdOut"), self.write_stdout
            )
        self.connect(
            self.environment, PYSIGNAL("writeStdErr"), self.write_stderr
            )
        
        left_layout = QVBoxLayout(left_frame)
        left_layout.addWidget(self.text_browser, 1)
        left_layout.addWidget(self.combo_edit, 0)
        
        layout = QHBoxLayout(central_widget)
        layout.addWidget(left_frame, 1)
        layout.addWidget(self.captured_widgets, 0)
        
        self.setCentralWidget(central_widget)
        
        # Print a welcome message.
        self.add_output(start_up)
    
    def setup_environment(self):
    
        # Create an environment to run Python scripts in.
        # After this, the self.environment object will be available for use.
        # This contains the module object which can have objects added to it.
        self.environment = Environment()
        
        # Redirect stdin, stdout and stderr to suitable objects.
        self.old_stdin = sys.__stdin__
        stdin = Buffer()
        stdin.read_callback = self.read_stdin
        sys.__stdin__ = sys.stdin = stdin
        
        self.old_stdout = sys.__stdout__
        stdout = Buffer()
        stdout.write_callback = self.write_stdout
        sys.__stdout__ = sys.stdout = stdout
        
        #self.old_stderr = sys.__stderr__
        #stderr = Buffer()
        #stderr.write_callback = self.write_stderr
        #sys.__stderr__ = sys.stderr = stderr
        
        # Create a completer for the environment.
        self.completer = Completer(self.environment.module)
        
        # Create a history of input strings and a variable to contain the
        # currently edited text.
        self.history = History()
        self.current_text = ""
        
        # Create a list of editors.
        self.editors = []
        
        # Widgets created in the user's namespace will be "captured" if the
        # following flag is set.
        self.capture_widgets = 0
        
        # Add a capture function to the environment's module to allow
        # arbitrary widgets to be added to the widget box.
        widget_box = self.captured_widgets
        
        def capture(widget):
        
            """capture(widget)
            
            Add a widget to the box of captured widgets.
            """
            if isinstance(widget, QWidget):
                widget_box.add_widgets((widget,))
            else:
                print "%s is not an instance of a QWidget subclass." % \
                    repr(widget)
        
        self.environment.module.capture = capture
        
        help_fn = self.help
        
        def docs(obj):
        
            if isinstance(obj, QObject):
            
                class_name = obj.__class__.__name__+".html"
                
                help_fn("class", class_name)
            
            elif type(obj) == types.BuiltinFunctionType:
            
                parent = obj.__self__
                
                if hasattr(parent, "__name__"):
                
                    class_name = parent.__name__
                
                elif hasattr(parent, "__class__"):
                
                    class_name = parent.__class__.__name__
                
                else:
                
                    return
                
                help_fn("function", class_name, obj.__name__)
            
            else:
            
                try:
                
                    if issubclass(obj, QObject):
                    
                        help_fn("class", obj.__name__)
                        return
                
                except TypeError:
                
                    pass
        
        self.environment.module.docs = docs
    
    def help(self, object_type, class_name, object_name = None):
    
        """help(self, object_type, class_name, object_name = None)
        
        Look for documentation in the Qt HTML documentation and show an
        excerpt if found.
        """
        
        if class_name[:3] == "qt.":
        
            class_name = class_name[3:]
        
        page_name = class_name.lower()+".html"
        page_path = os.path.join(self.documents_path, page_name)
        
        try:
        
            page = open(page_path).read()
        
        except IOError:
        
            return
        
        if object_type == "class":
        
            start = page.find("<h2>Detailed Description</h2>")
            
            if start != -1:
            
                end = page.find("<hr>", start)
                
                if end != -1:
                
                    self.add_output(page[start:end])
        
        elif object_type == "function":
        
            i = 0
            matches = []
            
            while i < len(page):
            
                start = page.find("<h3 class=fn", i)
                
                if start == -1: break
                
                end = page.find("</h3>", start)
                
                if end == -1: break
                
                found = page.find("::" + object_name, start, end)
                
                if found != -1:
                
                    next = page.find("<h3", end)
                    rule = page.find("<hr>", end)
                    
                    if next != -1 and next < rule:
                    
                        matches.append(page[start:next])
                    
                    elif rule != -1 and rule < next:
                    
                        matches.append(page[start:rule])
                
                i = end + 5
            
            self.add_output("<p>\n".join(matches))
    
    def setup_menus(self):
    
        # File menu
        self.file_menu = QPopupMenu(self)
        exit_action = QAction(
            self.tr("E&xit"), QKeySequence(self.tr("Ctrl+Q")), self
            )
        self.connect(exit_action, SIGNAL("activated()"), self, SLOT("close()"))
        exit_action.addTo(self.file_menu)
        
        # Environment menu
        self.environment_menu = QPopupMenu(self)
        self.environment_menu.setCheckable(1)
        
        editor_action = QAction(
            self.tr("Create edi&tor"), QKeySequence(self.tr("Ctrl+Shift+E")),
            self
            )
        self.capture_action = QAction(
            self.tr("&Capture widgets"), QKeySequence(), self
            )
        self.capture_action.setToggleAction(True)
        #self.environment_menu.setItemParameter(capture_id, capture_id)
        
        self.connect(editor_action, SIGNAL("activated()"), self.create_editor)
        self.connect(self.capture_action, SIGNAL("activated()"),
                     self.set_capture_widgets)
        
        editor_action.addTo(self.environment_menu)
        self.capture_action.addTo(self.environment_menu)
        
        # History menu (dynamically generated)
        self.history_menu = QPopupMenu(self)
        
        # Help menu
        self.help_menu = QPopupMenu(self)
        about_action = QAction(
            self.tr("&About %1").arg(appname), QKeySequence(), self
            )
        about_action.addTo(self.help_menu)
        
        self.connect(about_action, SIGNAL("activated()"), self.show_help)
        self.connect(
            self.history_menu, SIGNAL("aboutToShow()"), self.prepare_history
            )
        self.connect(
            self.history_menu, SIGNAL("activated(int)"), self.select_history
            )
        
        # Add the menus to the menu bar.
        self.menuBar().insertItem(self.tr("&File"), self.file_menu)
        self.menuBar().insertItem(
            self.tr("&Environment"), self.environment_menu
            )
        self.menuBar().insertItem(self.tr("H&istory"), self.history_menu)
        self.menuBar().insertSeparator()
        self.menuBar().insertItem(self.tr("&Help"), self.help_menu)
    
    def eventFilter(self, obj, event):
    
        """eventFilter(self, event)
        
        Monitor the events from the line edit widget.
        """
        
        if obj == self.combo_edit:
        
            if event.type() == QEvent.KeyPress:
            
                line_edit = self.combo_edit.lineEdit()
                if event.key() == Qt.Key_Tab:
                
                    self.complete_text(
                        str(line_edit.text()), line_edit.cursorPosition()
                        )
                    return 1
                
                elif event.key() == Qt.Key_Return:
                
                    self.execute_text(str(line_edit.text()))
                    return 1
                
                else:
                
                    self.remove_combo_items()
                    self.combo_edit.listBox().hide()
                    return 1
        
        elif obj == self.combo_edit.lineEdit():
        
            if event.type() == QEvent.KeyPress:
            
                if event.key() == Qt.Key_Up and \
                    self.combo_edit.listBox().count() == 0:
                
                    obj.setText(self.history.previous(str(obj.text())))
                    return 1
                
                elif event.key() == Qt.Key_Down and \
                    self.combo_edit.listBox().count() == 0:
                
                    obj.setText(self.history.next(str(obj.text())))
                    return 1
        
        return QMainWindow.eventFilter(self, obj, event)
    
    def create_editor(self):
    
        """Show an editor for Python code."""
        widget = Editor(self.environment)
        self.editors.append(widget)
        widget.show()
    
    def set_capture_widgets(self):
    
        self.capture_widgets = self.capture_action.isOn()
        
        if self.capture_widgets:
            self.captured_widgets.show()
        else:
            self.captured_widgets.hide()
    
    def prepare_history(self):
    
        self.history_menu.clear()
        
        current, items = self.history.items()
        for i in range(len(items)):
        
            item = items[i]
            item_id = self.history_menu.insertItem(item, i)
            
            if current == i:
                self.history_menu.setItemChecked(item_id, 1)
    
    def select_history(self, index):
    
        self.combo_edit.setCurrentText(
            self.history.select(index, str(self.combo_edit.currentText()))
            )
    
    def show_help(self):
    
        QMessageBox.about(
            self, self.tr("About %1").arg(appname),
            self.tr(
                "<p><b>An interactive editing environment for PyQt.</b>\n\n"
                "<p>Copyright (C) 2004-2005, David Boddie\n\n"
                "<p>Type peeq_license for license information."
                )
            )
    
    # Utility methods
    
    def remove_combo_items(self):
    
        # Unlike the clear method on its own, this will not leave the combo
        # box with a current item that is outside the range of items it thinks
        # are available.
        line_text = self.combo_edit.lineEdit().text()
        self.combo_edit.clear()
        self.combo_edit.setCurrentText(line_text)
    
    def execute_text(self, text):
    
        # Examine the text submitted.
        display_text = text = str(text)

        # Submit it to the user's running environment.
        try:

            # Retrieve a string value from the environment's execute method.
            if text[-1:] != ["\n"]:
            
                text = text + "\n"
            
            # Echo the text submitted to the output window.
            self.add_output(
                '<font color="blue"><tt>%s</tt></font>' % encode(display_text)
                )
            
            if self.capture_widgets:
                environment_names = dir(self.environment.module)
            
            # Execute the user's input.
            value = self.environment.execute(text)
            
            if self.capture_widgets:
                new_names = filter(
                    lambda name: name not in environment_names,
                    dir(self.environment.module)
                    )
                self.capture_objects(new_names)
            
            if isinstance(value, Traceback):

                self.write_value(value.text)

            elif value is not NoValue and value is not None:

                self.write_value(repr(value))
        
        except SyntaxError:
        
            pass
        
        else:
        
            # Clear the line edit and save the command in the command history.
            self.combo_edit.clear()
            self.history.append(display_text)
    
    def complete_text(self, text, cursor):
    
        start_at, matches = self.completer.match(text, cursor)
        
        if not matches:
            return
        
        line_text = self.combo_edit.lineEdit().text()
        self.combo_edit.clear()
        
        next = min(filter(
            lambda index: index != -1,
            map(lambda c: text.find(c, cursor), "(),. ") + [len(text)]
            ))
        
        if next == len(text):
            rest = ""
        else:
            rest = text[next:]
        
        if len(matches) == 1:
        
            self.combo_edit.setCurrentText(text[:start_at]+matches[0]+rest)
            self.combo_edit.lineEdit().setCursorPosition(
                start_at + len(matches[0])
                )
        
        elif len(matches) > 1:
        
            # Insert the item back into the list so that the line edit's text
            # doesn't get overwritten.
            self.combo_edit.insertItem(text)
            
            for match in matches:
            
                # Insert the combined text for each match into the combo box.
                self.combo_edit.insertItem(text[:start_at]+match+rest)
            
            self.combo_edit.popup()
            #self.combo_edit.lineEdit().setFocus()
    
    def capture_objects(self, names):
    
        objects = map(
            lambda name: self.environment.module.__dict__[name], names
            )
        
        widgets = filter(lambda obj: isinstance(obj, QWidget), objects)
        
        self.captured_widgets.add_widgets(widgets)
    
    # Input/output methods
    
    def add_output(self, output):
    
        self.text_browser.append(output)
    
    def read_stdin(self, buf, data):
    
        # Flush stdout and stderr before obtaining input.
        sys.stdout.flush()
        sys.stderr.flush()
        
        new_data, ok = QInputDialog.getText(
            appname, self.tr("Request for input"), QLineEdit.Normal,
            data, self
            )
        
        if ok == 1:
        
            # Echo stdin on stdout.
            sys.stdout.write(str(new_data)+"\n")
            
            return str(new_data)+"\n"
        
        else:
        
            return str(data)
    
    def process_output(self, buf, data, flush):
    
        # Take the data recently added to the buffer and add it to the
        # HTML content.
        
        if flush == 1 or data[-1:] == "\n":
        
            buf.seek(0, 0)
            
            # Use the public read method to extract the actual data.
            data = buf.read()
            
            #if flush == 0 and (data[:-1] == "\n" and data != "\n"):
            #
            #    # Remove trailing newlines from strings which contain
            #    # other characters, leaving single newlines alone.
            #    data = data[:-1]
            
            return self.wrap_text(data)
        
        return None
    
    def wrap_text(self, data):
    
        # Reformat, taking carriage return characters into account.
        lines = string.split(data, "\n")

        new_lines = []

        for line in lines:

            # Split the line at each carriage return encountered.
            words = string.split(line, "\r")

            # Construct a new line by overlaying all the "words"
            # which make up the line.
            length = reduce(lambda x,y: max(x, len(y)), words, 0)
            new_line = " " * min(self.display_width, length)

            for word in words:

                # Simply overlay the word over the current line string.
                # This is not optimal if the words are very long; in that
                # case they should be wrapped at a specified column number.
                new_line = word + new_line[len(word):]

            new_lines.append(new_line)

        data = string.join(new_lines, "\n")

        return encode(str(data), preformatted = 0)
    
    def write_value(self, text):
    
        output = self.wrap_text(text)
        
        self.add_output("<tt>%s</tt>\n" % output)
    
    def write_stdout(self, buf, data, flush = 0):
    
        output = self.process_output(buf, data, flush)
        
        if output is not None:
        
            # Use the appropriate formatter to present the value
            # obtained.
            
            self.add_output('<font color="#009900"><tt>%s</tt></font>' % output)
            
            # Clear the buffer.
            buf.clear()
    
    def write_stderr(self, buf, data, flush = 0):
    
        output = self.process_output(buf, data, flush)
        
        if output is not None:
        
            # Use the appropriate formatter to present the value
            # obtained.
            
            self.add_output('<font color="red"><tt>%s</tt></font>' % output)
            
            # Clear the buffer.
            buf.clear()


def main():

    # Find the path to this program.
    global app_path, app_file
    
    app_path, app_file = os.path.split(sys.argv[0])
    
    # If the working directory is not in the sys.path list then add it.
    
    pwd = os.getenv("PWD")
    
    if pwd != "" and pwd not in sys.path:
    
        sys.path.insert(0, pwd)
    
    app = QApplication(sys.argv)
    
    # The GUI interface object checks whether the command line has finished
    # and, if so, exits the application.
    gui = GUIInterface(app)
    gui.resize(640, 480)
    gui.show()
    app.setMainWidget(gui)
    
    # Wait for the Qt application to exit.
    return app.exec_loop()


if __name__ == "__main__":

    result = main()
    
    # Exit
    sys.exit(result)
